from http import HTTPStatus
from typing import Awaitable, Optional, Tuple

from ..content import ContentType
from .base import ASGIHeaders, ASGIResponse, Sender


HTTP_RESPONSE_START = 'http.response.start'
JSON_CONTENT_HEADER = (
    b'content-type',
    ContentType.APPLICATION_JSON.value.encode(),
)
HTML_CONTENT_HEADER = (b'content-type', ContentType.TEXT_HTML.value.encode())
PLAINTEXT_CONTENT_HEADER = (
    b'content-type',
    ContentType.TEXT_PLAIN.value.encode(),
)

JSON_RESPONSE: ASGIResponse = {
    'type': HTTP_RESPONSE_START,
    'status': HTTPStatus.OK.value,
    'headers': (JSON_CONTENT_HEADER,),
}

HTML_RESPONSE: ASGIResponse = {
    'type': HTTP_RESPONSE_START,
    'status': HTTPStatus.OK.value,
    'headers': (HTML_CONTENT_HEADER,),
}

PLAINTEXT_RESPONSE: ASGIResponse = {
    'type': HTTP_RESPONSE_START,
    'status': HTTPStatus.OK.value,
    'headers': (PLAINTEXT_CONTENT_HEADER,),
}

NOTFOUND_RESPONSE: ASGIResponse = {
    'type': HTTP_RESPONSE_START,
    'status': HTTPStatus.NOT_FOUND.value,
    'headers': (),
}

METHOD_NOT_ALLOWED_RESPONSE: ASGIResponse = {
    'type': HTTP_RESPONSE_START,
    'status': HTTPStatus.METHOD_NOT_ALLOWED.value,
    'headers': (),
}

NO_CONTENT_RESPONSE: ASGIResponse = {
    'type': HTTP_RESPONSE_START,
    'status': HTTPStatus.NO_CONTENT.value,
    'headers': (),
}


def make_response(
    content_length: Optional[int],
    status: HTTPStatus,
    headers: Optional[ASGIHeaders],
    default_value: ASGIResponse,
    default_content_header: Tuple[bytes, bytes],
) -> ASGIResponse:
    if content_length is None:
        return default_value

    default_headers: ASGIHeaders = (
        default_content_header,
        (b'content-length', str(content_length).encode()),
    )

    if headers:
        if isinstance(headers, tuple):
            headers = default_headers + headers
        else:
            headers = default_headers + tuple(headers)

    return {
        'type': HTTP_RESPONSE_START,
        'status': status.value,
        'headers': default_headers if not headers else headers,
    }


def make_json_response(
    content_length: Optional[int] = None,
    status: HTTPStatus = HTTPStatus.OK,
    headers: Optional[ASGIHeaders] = None,
) -> ASGIResponse:
    return make_response(
        content_length, status, headers, JSON_RESPONSE, JSON_CONTENT_HEADER
    )


def make_text_response(
    content_length: Optional[int] = None,
    status: HTTPStatus = HTTPStatus.OK,
    headers: Optional[ASGIHeaders] = None,
) -> ASGIResponse:
    return make_response(
        content_length,
        status,
        headers,
        PLAINTEXT_RESPONSE,
        PLAINTEXT_CONTENT_HEADER,
    )


def make_html_response(
    content_length: Optional[int] = None,
    status: HTTPStatus = HTTPStatus.OK,
    headers: Optional[ASGIHeaders] = None,
) -> ASGIResponse:
    return make_response(
        content_length, status, headers, HTML_RESPONSE, HTML_CONTENT_HEADER
    )


async def send_response(
    send: Sender, response: ASGIResponse, body: bytes
) -> None:
    await send(response)  # type: ignore
    await send(
        {'type': 'http.response.body', 'body': body, 'more_body': False}
    )


def send_not_found(send: Sender) -> Awaitable[None]:
    return send_response(send, NOTFOUND_RESPONSE, b'')


def send_method_not_allowed_response(send: Sender) -> Awaitable[None]:
    return send_response(send, METHOD_NOT_ALLOWED_RESPONSE, b'')



"""
ASGI App using dataclasses module for request/response objects
"""

__version__ = '0.12.0'

from apidaora.app import appdaora
from apidaora.bodies import GZipFactory
from apidaora.content import ContentType
from apidaora.exceptions import BadRequestError
from apidaora.header import Header
from apidaora.method import MethodType
from apidaora.responses import html, json, text
from apidaora.route.decorator import route


__all__ = [
    'BadRequestError',
    'GZipFactory',
    'ContentType',
    'MethodType',
    'appdaora',
    'Header',
    'html',
    'json',
    'text',
    'route',
]



from typing import Sequence, Union

from .asgi.app import asgi_app
from .asgi.base import ASGIApp
from .asgi.router import Controller, make_router
from .controllers.background_task import BackgroundTask


Controllers = Union[
    Controller,
    BackgroundTask,
    Sequence[Union[Controller, BackgroundTask]],
    Sequence[Controller],
    Sequence[BackgroundTask],
]


def appdaora(controllers: Controllers) -> ASGIApp:
    routes = []

    if not isinstance(controllers, Sequence):
        controllers = [controllers]

    for controller in controllers:
        if isinstance(controller, BackgroundTask):
            routes.append(controller.create.route)
            routes.append(controller.get_results.route)

        else:
            routes.append(controller.route)

    return asgi_app(make_router(routes))



import asyncio
import dataclasses
import datetime
import logging
import uuid
from concurrent.futures import ThreadPoolExecutor
from enum import Enum
from functools import partial
from http import HTTPStatus
from typing import Any, Callable, Coroutine, Dict, Type, TypedDict

import orjson
from jsondaora import as_typed_dict, jsondaora, typed_dict_asjson

from ..asgi.router import Controller
from ..exceptions import BadRequestError, InvalidTasksRepositoryError
from ..method import MethodType
from ..responses import Response, json
from ..route.factory import make_route


logger = logging.getLogger(__name__)


try:
    import aioredis
except Exception:
    aioredis = None


class TaskStatusType(Enum):
    RUNNING = 'running'
    FINISHED = 'finished'
    ERROR = 'error'


@jsondaora
class TaskInfo(TypedDict):
    task_id: str
    start_time: str
    status: str


@dataclasses.dataclass
class BackgroundTask:
    create: Controller
    get_results: Controller


def make_background_task(
    controller: Callable[..., Any],
    path_pattern: str,
    max_workers: int = 10,
    tasks_repository: Any = None,
) -> BackgroundTask:
    if asyncio.iscoroutinefunction(controller):
        logger.warning(
            'Async tasks can potentially block your application, use with care. '
            'It use is recommended just for small tasks or non-blocking operations.'
        )

    tasks_repository = get_tasks_repository(tasks_repository)

    annotations = getattr(controller, '__annotations__', {})
    result_annotation = annotations.get('return', str)

    @jsondaora
    class FinishedTaskInfo(TaskInfo):
        end_time: str
        result: result_annotation  # type: ignore

    create_task = make_create_task(
        controller, tasks_repository, FinishedTaskInfo, max_workers
    )
    get_task_results = make_get_task_results(
        tasks_repository, FinishedTaskInfo
    )

    create_task.__annotations__ = {
        name: type_ for name, type_ in annotations.items() if name != 'return'
    }

    return BackgroundTask(
        make_route(path_pattern, MethodType.POST, create_task).controller,
        make_route(path_pattern, MethodType.GET, get_task_results).controller,
    )


def get_iso_time() -> str:
    return (
        datetime.datetime.now()
        .replace(microsecond=0, tzinfo=datetime.timezone.utc)
        .isoformat()
    )


@dataclasses.dataclass
class BaseTasksRepository:
    async def set(
        self, key: Any, value: Any, task_cls: Type[Any] = TaskInfo
    ) -> None:
        ...

    async def get(self, key: Any, finished_task_cls: Type[Any]) -> Any:
        ...

    def close(self) -> None:
        ...


@dataclasses.dataclass
class SimpleTasksRepository(BaseTasksRepository):
    data_source: Dict[str, Any]

    async def set(
        self, key: Any, value: Any, task_cls: Type[Any] = TaskInfo
    ) -> None:
        self.data_source[key] = value

    async def get(self, key: Any, finished_task_cls: Type[Any]) -> Any:
        return self.data_source[key]


def get_tasks_repository(tasks_repository: Any) -> Any:
    if tasks_repository is None:
        return SimpleTasksRepository({})

    elif isinstance(tasks_repository, str) and tasks_repository.startswith(
        'redis://'
    ):
        if aioredis is None:
            raise InvalidTasksRepositoryError("'aioredis' package not found!")

        return partial(get_redis_tasks_repository, tasks_repository)

    elif isinstance(tasks_repository, BaseTasksRepository):
        return tasks_repository

    raise InvalidTasksRepositoryError(tasks_repository)


def make_create_task(
    controller: Callable[..., Any],
    tasks_repository_: Any,
    finished_task_info_cls: Any,
    max_workers: int,
) -> Callable[..., Coroutine[Any, Any, Response]]:
    executor = ThreadPoolExecutor(max_workers)

    async def create_task(*args: Any, **kwargs: Any) -> Response:
        task_id = uuid.uuid4()
        tasks_repository = tasks_repository_

        if asyncio.iscoroutinefunction(controller):
            if isinstance(tasks_repository, partial):
                tasks_repository = await tasks_repository()

            loop = asyncio.get_running_loop()
            wrapper = make_task_wrapper(
                tasks_repository,
                task_id,
                finished_task_info_cls,
                controller,
                *args,
                **kwargs,
            )
            asyncio.run_coroutine_threadsafe(wrapper(), loop)

        else:
            done_callback = make_done_callback(
                tasks_repository, task_id, finished_task_info_cls
            )

            future = executor.submit(controller, *args, **kwargs)
            future.add_done_callback(done_callback)

        start_time = get_iso_time()
        task = TaskInfo(
            task_id=str(task_id),
            start_time=start_time,
            status=TaskStatusType.RUNNING.value,
        )

        if isinstance(tasks_repository, partial):
            tasks_repository = await tasks_repository()

        await tasks_repository.set(
            task_id, task, task_cls=finished_task_info_cls
        )

        if not asyncio.iscoroutinefunction(controller):
            tasks_repository.close()

        return json(task, status=HTTPStatus.CREATED)

    return create_task


def make_task_wrapper(
    tasks_repository: Any,
    task_id: uuid.UUID,
    finished_task_info_cls: Any,
    controller: Callable[..., Any],
    *args: Any,
    **kwargs: Any,
) -> Callable[..., Coroutine[Any, Any, TaskInfo]]:
    async def wrapper() -> Any:
        result = await controller(*args, **kwargs)
        task = await tasks_repository.get(task_id, finished_task_info_cls)
        finished_task = finished_task_info_cls(
            end_time=get_iso_time(),
            result=result,
            status=TaskStatusType.FINISHED.value,
            task_id=task['task_id'],
            start_time=task['start_time'],
        )
        await tasks_repository.set(
            task_id, finished_task, task_cls=finished_task_info_cls
        )
        tasks_repository.close()

    return wrapper


def make_done_callback(
    tasks_repository_: Any, task_id: uuid.UUID, finished_task_info_cls: Any
) -> Callable[[Any], None]:
    def done_callback(future: Any) -> None:
        result = future.result()
        policy = asyncio.get_event_loop_policy()
        loop = policy.new_event_loop()
        tasks_repository = tasks_repository_

        if isinstance(tasks_repository, partial):
            tasks_repository = loop.run_until_complete(tasks_repository())

        task = loop.run_until_complete(
            tasks_repository.get(task_id, finished_task_info_cls)
        )
        finished_task = finished_task_info_cls(
            end_time=get_iso_time(),
            result=result,
            status=TaskStatusType.FINISHED.value,
            task_id=task['task_id'],
            start_time=task['start_time'],
        )
        loop.run_until_complete(
            tasks_repository.set(
                task_id, finished_task, task_cls=finished_task_info_cls
            )
        )
        tasks_repository.close()

    return done_callback


def make_get_task_results(
    tasks_repository_: Any, finished_task_info_cls: Any
) -> Callable[..., Coroutine[Any, Any, TaskInfo]]:
    async def get_task_results(task_id: str) -> finished_task_info_cls:  # type: ignore
        tasks_repository = tasks_repository_

        if isinstance(tasks_repository, partial):
            tasks_repository = await tasks_repository()

        try:
            task = await tasks_repository.get(
                uuid.UUID(task_id), finished_task_info_cls
            )
            tasks_repository.close()
            return task  # type: ignore
        except KeyError:
            raise BadRequestError(
                name='invalid_task_id', info={'task_id': task_id}
            )
        except ValueError as error:
            if error.args == ('badly formed hexadecimal UUID string',):
                raise BadRequestError(
                    name='invalid_task_id', info={'task_id': task_id}
                )

            raise error from None

    return get_task_results


if aioredis is not None:

    @dataclasses.dataclass
    class RedisTasksRepository(BaseTasksRepository):
        data_source: aioredis.Redis

        async def set(
            self, key: Any, value: Any, task_cls: Type[Any] = TaskInfo
        ) -> None:
            await self.data_source.set(
                str(key), typed_dict_asjson(value, task_cls)
            )

        async def get(self, key: Any, finished_task_cls: Type[Any]) -> Any:
            value = await self.data_source.get(str(key))

            if value:
                value = orjson.loads(value)

                if value['status'] == TaskStatusType.RUNNING.value:
                    return as_typed_dict(value, TaskInfo)

                if value['status'] == TaskStatusType.FINISHED.value:
                    return as_typed_dict(value, finished_task_cls)

            raise KeyError(key)

        def close(self) -> None:
            self.data_source.close()

    async def get_redis_tasks_repository(uri: str) -> RedisTasksRepository:
        data_source = await aioredis.create_redis_pool(uri)
        return RedisTasksRepository(data_source)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['apidaora',
 'apidaora.asgi',
 'apidaora.controllers',
 'apidaora.core',
 'apidaora.route']

package_data = \
{'': ['*']}

install_requires = \
['jsondaora', 'dictdaora']

extras_require = \
{'doc': ['mkdocs', 'mkdocs-material', 'markdown-include'],
 'redis': ['aioredis'],
 'test': ['asgi-testclient',
          'bumpversion',
          'black',
          'flake8',
          'isort',
          'ipython',
          'mypy',
          'pytest-asyncio',
          'pytest-cov',
          'pytest-mock',
          'pytest>=5.1.1',
          'uvicorn',
          'aioredis']}

setup(name='apidaora',
      version='0.12.0',
      description='ASGI App using dataclasses module for request/response objects',
      author='Diogo Dutra',
      author_email='diogodutradamata@gmail.com',
      url='https://github.com/dutradda/apidaora',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.8',
     )



from http import HTTPStatus
from typing import Dict

from jsondaora import jsondaora

from apidaora import BadRequestError, Header, appdaora, json, route


# Domain layer, here are the domain related definitions
# it is apidaora/framework/http independent


@jsondaora
class You:
    name: str
    last_name: str
    age: int


DB: Dict[str, You] = {}


def add_you(you):
    if you.name in DB:
        raise YouAlreadyBeenAddedError(you.name)
    DB[you.name] = you


def get_you(name):
    try:
        return DB[name]
    except KeyError:
        raise YouWereNotFoundError(name)


class DBError(Exception):
    @property
    def info(self):
        return {'name': self.args[0]}


class YouAlreadyBeenAddedError(DBError):
    name = 'you-already-been-added'


class YouWereNotFoundError(DBError):
    name = 'you-were-not-found'


# Application layer, here are the http related definitions

# See: https://dutrdda.github.io/apidaora/tutorial/headers/
class ReqID(Header, type=str, http_name='http_req_id'):
    ...


@route.post('/you/')
async def add_you_controller(req_id: ReqID, body: You):
    try:
        add_you(body)
    except YouAlreadyBeenAddedError as error:
        raise BadRequestError(name=error.name, info=error.info) from error

    return json(body, HTTPStatus.CREATED, headers=(req_id,))


@route.get('/you/{name}')
async def get_you_controller(name: str, req_id: ReqID):
    try:
        return json(get_you(name), headers=(req_id,))
    except YouWereNotFoundError as error:
        raise BadRequestError(name=error.name, info=error.info) from error


app = appdaora([add_you_controller, get_you_controller])
